@@ -3,9 +3,12 @@ import pathlib
 import ast
 import json
 import typing
+import time
+
 from .function import gather_function_attributes, FunctionAttributes
-from dataclasses import dataclass
-from types import BuiltinFunctionType
+import dataclasses
+from dataclasses import dataclass, is_dataclass, fields
+from types import BuiltinFunctionType, FrameType
 from typing import Dict, List, Callable, Mapping, Any, Union, Iterable, Optional, Type
 
 from aws_cdk import (
@@ -17,7 +20,7 @@ from aws_cdk import (
 from aws_cdk.aws_stepfunctions import JsonPath
 
 from .condition import build_condition
-from .s3 import write_json, read_json, build_s3_write_json_step, build_s3_read_json_step
+from .service_operations import service_operations
 
 SFN_INDEX = 0
 
@@ -41,6 +44,10 @@ def concurrent(iterable, max_concurrency: Optional[int] = None):
     pass
 
 
+def parallel(threads):
+    pass
+
+
 def event(func: Callable):
     pass
 
@@ -61,10 +68,20 @@ def await_token(
     pass
 
 
+@dataclass
+class SFNError(Exception):
+    error: str
+    cause: str = ""
+
+    @classmethod
+    def error(cls, error: str):
+        return cls(error)
+
+
 def state_machine(
     cdk_stack: Stack,
     sfn_name: str,
-    local_values,
+    local_values=None,  # DEPRECATED
     express=False,
     skip_pass=True,
     return_vars: Optional[Union[List[str], Mapping[str, typing.Type]]] = None,
@@ -76,17 +93,27 @@ def state_machine(
 
     def decorator(func):
         func_attrs = gather_function_attributes(func, None, return_vars)
-        fts = FunctionToSteps(cdk_stack, func_attrs, local_values, skip_pass=skip_pass)
-        func.state_machine = sfn.StateMachine(
-            cdk_stack,
-            sfn_name,
-            state_machine_name=sfn_name,
-            state_machine_type=sfn.StateMachineType.EXPRESS
-            if express
-            else sfn.StateMachineType.STANDARD,
-            definition=fts.build_sfn_definition(),
-        )
-        func.output = func_attrs.output
+        stack = inspect.stack()
+        try:
+            fts = FunctionToSteps(
+                cdk_stack, func_attrs, stack[1].frame, skip_pass=skip_pass
+            )
+            print(sfn_name)
+            func.state_machine = sfn.StateMachine(
+                cdk_stack,
+                sfn_name,
+                state_machine_name=sfn_name,
+                state_machine_type=sfn.StateMachineType.EXPRESS
+                if express
+                else sfn.StateMachineType.STANDARD,
+                definition=fts.build_sfn_definition()[0],
+            )
+            func.output = func_attrs.output
+            if fts.additional_policies:
+                for policy in fts.additional_policies:
+                    func.state_machine.add_to_role_policy(policy)
+        finally:
+            del stack
         return func
 
     return decorator
@@ -97,17 +124,19 @@ class FunctionToSteps:
         self,
         cdk_stack: Stack,
         func_attrs: FunctionAttributes,
-        local_values,
+        frame: FrameType,
         skip_pass=True,
     ):
         global SFN_INDEX
         self.cdk_stack = cdk_stack
         self.func = func_attrs.func
         self.output = func_attrs.output
-        self.local_values = local_values
+        self.local_values = frame.f_locals
+        self.global_values = frame.f_globals
         self.state_number = 0
         self.sfn_number = SFN_INDEX
         self.skip_pass = skip_pass
+        self.additional_policies = []
         SFN_INDEX += 1
 
         self.ast = func_attrs.tree
@@ -141,10 +170,15 @@ class FunctionToSteps:
         advance(next_, c, n)
 
         write_definition_json(self.func.__name__, start)
-        return start
+        return start, next_
+
+    def get_frame_value(self, name):
+        return self.local_values.get(name, self.global_values.get(name))
 
 
 class SFNScope:
+    # TODO: Fix all of the function call handling which largely ignores keyword args
+
     def __init__(self, fts: FunctionToSteps):
         self.fts = fts
         self.cdk_stack = fts.cdk_stack
@@ -167,7 +201,9 @@ class SFNScope:
         next_ = advance(start.next, c, n)
         return start, next_
 
-    def build_register_assignment(self, values: Dict, register_path: str = ""):
+    def build_register_assignment(
+        self, values: Dict, register_path: str = "", value_types=None
+    ):
         params = values.copy()
         for k, v in params.items():
             self._updated_var(k)
@@ -181,7 +217,12 @@ class SFNScope:
         for key in values.keys():
             if key not in self.variables:
                 # TODO: Assign the type of the value
-                self.add_var(key)
+                self.add_var(
+                    key,
+                    var_type=value_types.get(key, typing.Any)
+                    if value_types
+                    else typing.Any,
+                )
 
         params = {update_param_name(k, v): v for k, v in params.items()}
         return params
@@ -200,6 +241,13 @@ class SFNScope:
     def handle_body(self, body: List[ast.stmt]) -> (List[sfn.IChainable], Callable):
         chain = []
         next_ = None
+        # If the first statement is a function doc string we can drop it
+        if (
+            len(body) > 0
+            and isinstance(body[0], ast.Expr)
+            and isinstance(body[0].value, ast.Constant)
+        ):
+            body = body[1:]
         for stmt in body:
             c, n = self.handle_op(stmt)
             chain.extend(c)
@@ -213,13 +261,48 @@ class SFNScope:
         if isinstance(stmt, ast.AnnAssign) or isinstance(stmt, ast.Assign):
             # if isinstance(stmt.value, ast.Constant):
             #    return self.handle_assign_value(stmt)
-            if isinstance(stmt.value, ast.Call) and isinstance(
-                stmt.value.func, ast.Name
-            ):
-                return self.handle_call_function(stmt.value, stmt)
+            # TODO: Revisit this hack for dataclasses
+            if isinstance(stmt.value, ast.Call):
+                call = stmt.value
+                if isinstance(call.func, ast.Name):
+                    func = self.fts.get_frame_value(call.func.id)
+
+                    if is_dataclass(func):
+                        dc_fields = fields(func)
+                        # print(f"Dataclass {func}")
+                        # print("fields:")
+                        # print([f.name for f in dc_fields])
+                        if len(call.args) == 1:
+                            arg = call.args[0]
+                            if isinstance(arg, ast.Starred):
+                                if isinstance(arg.value, ast.Call):
+                                    return self.handle_call_function(
+                                        arg.value, stmt, func
+                                    )
+                        value = self.build_dataclass_default_structure(func)
+                        for key, val in zip(value.keys(), call.args):
+                            value[key] = val
+                        for key in call.keywords:
+                            value[key.arg] = self.generate_value_repr(key.value)
+                        self.validate_dataclass_values(value)
+                        # print("Mapped object")
+                        # print(value)
+                        # TODO Fix this hack that assumes a non-annotated assignment
+                        target = stmt.targets[0].id
+                        # print(self.build_register_assignment({target: value}))
+                        assign = sfn.Pass(
+                            self.cdk_stack,
+                            self.state_name(f"Assign {target}"),
+                            input_path="$.register",
+                            result_path="$.register",
+                            parameters=self.build_register_assignment({target: value}),
+                        )
+                        return [assign], assign.next
+                    else:
+                        return self.handle_call_function(call, stmt)
             if isinstance(stmt.value, ast.ListComp):
                 return self.handle_list_comp(stmt)
-            else:
+            elif stmt.value:
                 return self.handle_assign_value(stmt)
         elif isinstance(stmt, ast.If):
             return self.handle_if(stmt)
@@ -236,6 +319,8 @@ class SFNScope:
                     pass
         elif isinstance(stmt, ast.With):
             return self.handle_with(stmt)
+        elif isinstance(stmt, ast.While):
+            return self.handle_while(stmt)
         elif isinstance(stmt, ast.Try):
             return self.handle_try(stmt)
         elif isinstance(stmt, ast.For):
@@ -250,6 +335,30 @@ class SFNScope:
             else:
                 pass_step = sfn.Pass(self.cdk_stack, self.state_name("Pass"))
                 return [pass_step], pass_step.next
+        elif isinstance(stmt, ast.Raise):
+            if (
+                isinstance(stmt.exc, ast.Call)
+                and isinstance(stmt.exc.func, ast.Name)
+                and stmt.exc.func.id == "SFNError"
+                and len(stmt.exc.args) >= 1
+            ):
+                args = [self.map_arg(a, "register.") for a in stmt.exc.args]
+                fail_step = sfn.Fail(
+                    self.cdk_stack,
+                    self.state_name("Raise error"),
+                    error=args[0],
+                    cause=args[1] if len(args) > 1 else None,
+                )
+                return [fail_step], None
+            elif isinstance(stmt.exc, ast.Name):
+                prefix = f"$.register.{stmt.exc.id}"
+                fail_step = sfn.Fail(
+                    self.cdk_stack,
+                    self.state_name("Raise error"),
+                    error=f"{prefix}.Error",
+                    cause=f"{prefix}.Cause",
+                )
+                return [fail_step], None
 
         # Treat unhandled statements as a no-op
         print(f"Unhandled {repr(stmt)}")
@@ -304,6 +413,15 @@ class SFNScope:
     ):
         if isinstance(handler.type, ast.Name) and handler.type.id == "Exception":
             return CatchHandler(["States.ALL"], chain, result_path=result_path)
+        elif (
+            isinstance(handler.type, ast.Call)
+            and isinstance(handler.type.func, ast.Attribute)
+            and isinstance(handler.type.func.value, ast.Name)
+            and handler.type.func.value.id == "SFNError"
+        ):
+            if all([isinstance(a, ast.Constant) for a in handler.type.args]):
+                args = [a.value for a in handler.type.args]
+                return CatchHandler(args, chain, result_path=result_path)
         raise Exception(f"Unhandled exception of type {handler.type}")
 
     @staticmethod
@@ -340,7 +458,10 @@ class SFNScope:
             and isinstance(arg.value, ast.Name)
             and isinstance(arg.slice, ast.Constant)
         ):
-            return f"$.{var_path}{arg.value.id}[{arg.slice.value}]"
+            if isinstance(arg.slice.value, int):
+                return f"$.{var_path}{arg.value.id}[{arg.slice.value}]"
+            else:
+                return f"$.{var_path}{arg.value.id}.{arg.slice.value}"
         else:
             raise Exception("Args must be Name or Constant")
 
@@ -349,6 +470,7 @@ class SFNScope:
         # TODO: Support callable iterator
         max_concurrency = 1
         iterator_step = None
+        has_index = False
 
         # if the iterator is wrapped in a 'concurrent' function, capture the parameter as the concurrency for the map
         if (
@@ -373,6 +495,17 @@ class SFNScope:
             )
             iter_var = iterator.func.id
             items_path = f"$.iter.{iter_var}"
+        elif (
+            isinstance(iterator, ast.Call)
+            and isinstance(iterator.func, ast.Name)
+            and iterator.func.id == "enumerate"
+        ):
+            if len(iterator.args) == 1 and isinstance(iterator.args[0], ast.Name):
+                iter_var = iterator.args[0].id
+                items_path = f"$.register.{iter_var}"
+                has_index = True
+            else:
+                raise Exception("enumerate can only be used with variable values")
         elif isinstance(iterator, ast.Call) and isinstance(iterator.func, ast.Name):
             (
                 iterator_step,
@@ -389,31 +522,70 @@ class SFNScope:
             items_path = f"$.register.{iter_var}"
         else:
             raise Exception("Unsupported for-loop iterator, variables only")
-        return iter_var, items_path, iterator_step, max_concurrency
+        return iter_var, items_path, iterator_step, max_concurrency, has_index
+
+    def handle_while(self, stmt: ast.While):
+        condition, name = build_condition(stmt.test)
+        choice = sfn.Choice(
+            self.cdk_stack, self.state_name(name.replace("If", "While"))
+        )
+
+        def if_next(step):
+            choice.when(condition, step)
+
+        if_c, if_n = ChildScope(self).handle_body(stmt.body)
+        if stmt.orelse:
+            raise Exception("While else isn't currently supported")
+        if_n = advance(if_next, if_c, if_n)
+        advance(if_n, choice, None)
+        chain = [choice]
+        if if_c:
+            chain.extend(if_c)
+        return chain, [choice.otherwise]
 
     def handle_for(self, stmt: ast.For):
-        iter_var, items_path, iterator_step, max_concurrency = self._get_iterator(
-            stmt.iter
-        )
-        if not isinstance(stmt.target, ast.Name):
+        # print(ast.dump(stmt))
+        # print()
+        (
+            iter_var,
+            items_path,
+            iterator_step,
+            max_concurrency,
+            has_index,
+        ) = self._get_iterator(stmt.iter)
+        index_name = None
+        if (
+            isinstance(stmt.target, ast.Tuple)
+            and len(stmt.target.elts) == 2
+            and has_index
+        ):
+            index_name = stmt.target.elts[0].id
+            target_name = stmt.target.elts[1].id
+        elif isinstance(stmt.target, ast.Name):
+            target_name = stmt.target.id
+        else:
             raise Exception("Unsupported for-loop target, variables only")
 
+        map_parameters = {
+            "register.$": f"$.register",
+            f"{target_name}.$": "$$.Map.Item.Value",
+        }
+        if index_name:
+            map_parameters[f"{index_name}.$"] = "$$.Map.Item.Index"
+
         map_state = sfn.Map(
             self.cdk_stack,
             self.state_name(f"For {iter_var}"),
             max_concurrency=max_concurrency,
             items_path=items_path,
-            parameters={
-                "register.$": f"$.register",
-                f"{stmt.target.id}.$": "$$.Map.Item.Value",
-            },
+            parameters=map_parameters,
             result_path="$.register.loopResult",
         )
 
         # Create a scope for the for loop contents and build the contained steps
         # Also build an entry step to capture the iterator target
         map_scope = MapScope(self)
-        entry_step = map_scope.build_entry_step(stmt.target.id)
+        entry_step = map_scope.build_entry_step(target_name, index_name)
         chain, map_next_ = map_scope.handle_body(stmt.body)
         entry_step.next(chain[0])
         map_state.iterator(entry_step)
@@ -564,6 +736,13 @@ class SFNScope:
         elif isinstance(arg, ast.Constant):
             value = arg.value
             path_to_add = arg.value
+        elif (
+            isinstance(arg, ast.Subscript)
+            and isinstance(arg.value, ast.Name)
+            and isinstance(arg.slice, ast.Constant)
+        ):
+            value = f"{arg.value.id}.{arg.slice.value}"
+            path_to_add = f"$.register.{value}"
         else:
             raise Exception(f"Unexpected type {type(arg)} for list append")
         list_step = sfn.Pass(
@@ -592,9 +771,13 @@ class SFNScope:
         self, stmt: Union[ast.AnnAssign, ast.Assign]
     ) -> (List[sfn.IChainable], Callable):
         sub_target = None
+        assignment_type = typing.Any
         if isinstance(stmt, ast.AnnAssign):
             if isinstance(stmt.target, ast.Name):
                 var_name = stmt.target.id
+                # TODO: Handle subscripted annotations such as typing.List
+                if isinstance(stmt.annotation, ast.Name):
+                    assignment_type = self.fts.get_frame_value(stmt.annotation.id)
             elif (
                 isinstance(stmt.target, ast.Subscript)
                 and isinstance(stmt.target.value, ast.Name)
@@ -607,17 +790,28 @@ class SFNScope:
                     f"Unexpected assignment target of type {type(stmt.target)}"
                 )
         else:
-            if len(stmt.targets) == 1 and isinstance(stmt.targets[0], ast.Name):
-                var_name = stmt.targets[0].id
-            elif (
-                isinstance(stmt.targets[0], ast.Subscript)
-                and isinstance(stmt.targets[0].value, ast.Name)
-                and isinstance(stmt.targets[0].slice, ast.Constant)
-            ):
-                var_name = stmt.targets[0].value.id
-                sub_target = stmt.targets[0].slice.value
+            if len(stmt.targets) == 1:
+                target = stmt.targets[0]
+                if isinstance(target, ast.Name):
+                    var_name = target.id
+                elif (
+                    isinstance(target, ast.Subscript)
+                    and isinstance(target.value, ast.Name)
+                    and isinstance(target.slice, ast.Constant)
+                ):
+                    var_name = target.value.id
+                    sub_target = target.slice.value
+
+                # dataclass attribute assignment
+                elif isinstance(target, ast.Attribute) and isinstance(
+                    target.value, ast.Name
+                ):
+                    var_name = target.value.id
+                    sub_target = target.attr
+                else:
+                    raise Exception("Unexpected assignment")
             else:
-                raise Exception("Unexpected assignment")
+                raise Exception("More targets than expected")
         if sub_target:
             prep = sfn.Pass(
                 self.cdk_stack,
@@ -633,11 +827,50 @@ class SFNScope:
                 result_path="$.register",
                 parameters=self.build_register_assignment(
                     # {f"{var_name}": JsonPath.json_merge(f"$.{var_name}", "$.itm")}
-                    {f"{var_name}": f"States.JsonMerge($.{var_name}, $.itm, false)"}
+                    {f"{var_name}": f"States.JsonMerge($.{var_name}, $.itm, false)"},
+                    value_types={var_name: assignment_type},
                 ),
             )
             prep.next(assign)
             return [prep, assign], assign.next
+
+        # Handling for array addition
+        elif isinstance(stmt.value, ast.BinOp) and isinstance(stmt.value.op, ast.Add):
+            op = stmt.value
+            if isinstance(op.left, ast.Name):
+                left_name = op.left.id
+                left_type = self.variables.get(op.left.id)
+            else:
+                raise Exception("Unexpected left value in value addition")
+            if isinstance(op.right, ast.Name):
+                right_name = op.right.id
+                right_type = self.variables.get(op.right.id)
+            else:
+                raise Exception("Unexpected right value in value addition")
+            if left_type == typing.List and right_type == typing.List:
+                list_step = sfn.Pass(
+                    self.cdk_stack,
+                    self.state_name(f"Add {left_name} and {right_name}"),
+                    result_path="$.meta",
+                    parameters={
+                        "arrayConcat": JsonPath.string_at(
+                            f"States.Array($.register.{left_name}, $.register.{right_name})"
+                        )
+                    },
+                )
+                flatten_step = sfn.Pass(
+                    self.cdk_stack,
+                    self.state_name(f"Flatten {var_name}"),
+                    result_path="$.register",
+                    parameters=self.build_register_assignment(
+                        {var_name: JsonPath.string_at("$.meta.arrayConcat[*][*]")},
+                        "register.",
+                    ),
+                )
+                list_step.next(flatten_step)
+                return [list_step, flatten_step], flatten_step.next
+            else:
+                raise Exception("Addition is currently only supported for arrays")
         else:
             value = self.generate_value_repr(stmt.value)
             assign = sfn.Pass(
@@ -645,7 +878,9 @@ class SFNScope:
                 self.state_name(f"Assign {var_name}"),
                 input_path="$.register",
                 result_path="$.register",
-                parameters=self.build_register_assignment({var_name: value}),
+                parameters=self.build_register_assignment(
+                    {var_name: value}, value_types={var_name: assignment_type}
+                ),
             )
             return [assign], assign.next
 
@@ -723,29 +958,23 @@ class SFNScope:
     ) -> (sfn.State, List[str], str):
         if isinstance(call.func, ast.Name):
             # Get the function
-            func = self.fts.local_values.get(call.func.id)
+            func = self.fts.get_frame_value(call.func.id)
             result_prefix = ""
 
             # Build the parameters
             if func:
-                params = self.build_parameters(call, func)
-                if hasattr(func, "get_additional_params"):
-                    params.update(func.get_additional_params())
-            elif call.func.id == write_json.__name__:
-                params = self.build_parameters(call, write_json, False)
-            elif call.func.id == read_json.__name__:
-                params = self.build_parameters(call, read_json, False)
-            elif call.func.id in [
-                "time.sleep",
-                "sleep",
-                event.__name__,
-                await_token.__name__,
-            ]:
-                params = {}
+                if func in service_operations:
+                    params = self.build_parameters(call, func, False)
+                elif func in [time.sleep, event, await_token]:
+                    params = {}
+                else:
+                    params = self.build_parameters(call, func)
+                    if hasattr(func, "get_additional_params"):
+                        params.update(func.get_additional_params())
             else:
                 raise Exception(f"Unable to find function {call.func.id}")
 
-            if call.func.id in ["time.sleep", "sleep"]:
+            if func == time.sleep:
                 invoke = sfn.Wait(
                     self.cdk_stack,
                     self.state_name("Wait"),
@@ -753,7 +982,7 @@ class SFNScope:
                 )
                 return_vars = []
             elif (
-                call.func.id == event.__name__
+                func == event
                 and len(call.args) > 0
                 and isinstance(call.args[0], ast.Call)
             ):
@@ -761,7 +990,7 @@ class SFNScope:
                     call.args[0], result_path=result_path, invoke_event_=True
                 )
             elif (
-                call.func.id == await_token.__name__
+                func == await_token
                 and len(call.args) >= 2
                 and isinstance(call.args[0], ast.Call)
             ):
@@ -823,19 +1052,23 @@ class SFNScope:
                 )
                 return_vars = list(func.output.keys())
                 result_prefix = ".Output"
-            elif call.func.id == write_json.__name__:
-                invoke = build_s3_write_json_step(
-                    self.cdk_stack, self.state_name("S3 Write JSON"), **params
+            elif func in service_operations:
+                invoke = func.builder(
+                    self.cdk_stack, self.state_name(func.step_name), **params
                 )
-                return_vars = ["ETag"]
-                result_prefix = ""
-            elif call.func.id == read_json.__name__:
-                invoke = build_s3_read_json_step(
-                    self.cdk_stack, self.state_name("S3 Read JSON"), **params
-                )
-                return_vars = ["Body", "LastModified", "ETag"]
+                if hasattr(func, "additional_policies") and func.additional_policies:
+                    self.fts.additional_policies.extend(func.additional_policies)
+                return_vars = func.return_vars
                 result_prefix = ""
             else:
+                func_attrs = gather_function_attributes(func)
+                func_tree = func_attrs.tree.body[0]
+                with open(
+                    pathlib.Path("build", f"{func_attrs.name}_ast.txt"), "w"
+                ) as fp:
+                    fp.write(ast.dump(func_tree, indent=2))
+                chain, n = ChildScope(self).handle_body(func_tree.body)
+
                 raise Exception(
                     f"Function without an associated Lambda: {call.func.id}"
                 )
@@ -846,7 +1079,10 @@ class SFNScope:
             )
 
     def handle_call_function(
-        self, call: ast.Call, assign: Union[ast.Assign, ast.AnnAssign] = None,
+        self,
+        call: ast.Call,
+        assign: Union[ast.Assign, ast.AnnAssign] = None,
+        dataclass_=None,
     ) -> (List[sfn.IChainable], Callable):
         if self._is_intrinsic_function(call):
             if not assign:
@@ -860,7 +1096,9 @@ class SFNScope:
             if isinstance(target, ast.Name):
                 result_target = target.id
             else:
-                raise Exception("Invalid intrinsic function target")
+                raise Exception(
+                    f"Invalid intrinsic function target {target} for call {call.func.id}"
+                )
 
             register = sfn.Pass(
                 self.cdk_stack,
@@ -876,6 +1114,8 @@ class SFNScope:
             )
             chain = [invoke]
             next_ = invoke.next
+            context_target = None
+            CONTEXT_TARGET_NAMES = {"__execution_arn": "ExecutionArn"}
 
             if assign:
                 # Get the result variable names
@@ -890,28 +1130,56 @@ class SFNScope:
                     isinstance(t, ast.Name) for t in target.elts
                 ):
                     result_targets = [n.id for n in target.elts]
+                    if result_targets[0] in CONTEXT_TARGET_NAMES:
+                        context_target = result_targets[0]
+                        result_targets = result_targets[1:]
                 else:
                     raise Exception(
                         f"Unexpected result target of type {type(assign.target)}"
                     )
-                result_params = {
-                    v: JsonPath.string_at(f"$.out{result_prefix}.{r}")
-                    for v, r in zip(result_targets, return_vars)
-                }
-                if len(result_params) < len(result_targets):
-                    raise Exception(
-                        f"Unable to map all response targets to return values for {name}"
+                if dataclass_:
+                    # print("dataclass assignment")
+                    value = self.build_dataclass_default_structure(dataclass_)
+                    for key, val in zip(value.keys(), return_vars):
+                        value[key] = JsonPath.string_at(f"$.out.{val}")
+                    self.validate_dataclass_values(value)
+                    # print("Mapped object")
+                    # print(value)
+                    # TODO Fix this hack that assumes a non-annotated assignment
+                    # print(self.build_register_assignment({target.id: value}))
+                    register = sfn.Pass(
+                        self.cdk_stack,
+                        self.state_name(f"Register {call.func.id}"),
+                        input_path="$.register",
+                        result_path="$.register",
+                        parameters=self.build_register_assignment({target.id: value}),
                     )
-                register = sfn.Pass(
-                    self.cdk_stack,
-                    self.state_name(f"Register {call.func.id}"),
-                    input_path="$.register",
-                    result_path="$.register",
-                    parameters=self.build_register_assignment(result_params),
-                )
-                invoke.next(register)
-                chain.append(register)
-                next_ = register.next
+                    invoke.next(register)
+                    chain.append(register)
+                    next_ = register.next
+                else:
+                    result_params = {
+                        v: JsonPath.string_at(f"$.out{result_prefix}.{r}")
+                        for v, r in zip(result_targets, return_vars)
+                    }
+                    if context_target:
+                        result_params[context_target] = JsonPath.string_at(
+                            f"$.out.{CONTEXT_TARGET_NAMES[context_target]}"
+                        )
+                    if len(result_params) < len(result_targets):
+                        raise Exception(
+                            f"Unable to map all response targets to return values for {name}"
+                        )
+                    register = sfn.Pass(
+                        self.cdk_stack,
+                        self.state_name(f"Register {call.func.id}"),
+                        input_path="$.register",
+                        result_path="$.register",
+                        parameters=self.build_register_assignment(result_params),
+                    )
+                    invoke.next(register)
+                    chain.append(register)
+                    next_ = register.next
 
             return chain, next_
 
@@ -923,6 +1191,14 @@ class SFNScope:
         elif isinstance(value, ast.Call) and self._is_intrinsic_function(value):
             val, name = self._intrinsic_function(value)
             return val
+        elif (
+            isinstance(value, ast.Subscript)
+            and isinstance(value.value, ast.Name)
+            and isinstance(value.slice, ast.Constant)
+        ):
+            return JsonPath.string_at(
+                f"$.register.{value.value.id}.{value.slice.value}"
+            )
         else:
             raise Exception(f"Unanticipated return type: {value}")
 
@@ -1019,7 +1295,8 @@ class SFNScope:
             expr = f"$.{arg_value.id}"
             return JsonPath.string_at(expr) if gen_jsonpath else expr
         elif isinstance(arg_value, ast.Constant):
-            return arg_value.value
+            # TODO: Investigate this more, it appears that values of None are getting dropped altogether rather than using null
+            return arg_value.value if arg_value.value is not None else ""
         elif isinstance(arg_value, ast.List):
             expr = [self.generate_value_repr(val, False) for val in arg_value.elts]
             return JsonPath.array(*expr) if gen_jsonpath else expr
@@ -1071,7 +1348,7 @@ class SFNScope:
             and isinstance(arg_value.value, ast.Name)
             and arg_value.value.id == "self"
         ):
-            s = self.fts.local_values.get(arg_value.value.id)
+            s = self.fts.get_frame_value(arg_value.value.id)
             var = s.__getattribute__(arg_value.attr)
             return var
         elif isinstance(arg_value, ast.Call) and (
@@ -1088,6 +1365,27 @@ class SFNScope:
             print(ast.dump(arg_value, indent=2))
             raise Exception(f"Unexpected argument: {ast.dump(arg_value)}")
 
+    def build_dataclass_default_structure(self, dc):
+        dc_fields = dataclasses.fields(dc)
+        value = {}
+        for f in dc_fields:
+            if f.default is dataclasses.MISSING:
+                if f.default_factory is dataclasses.MISSING:
+                    value[f.name] = Ellipsis
+                elif f.default_factory is list:
+                    value[f.name] = []
+                else:
+                    raise Exception("Unsupported dataclass default factory")
+            else:
+                value[f.name] = self.generate_value_repr(ast.Constant(f.default))
+        return value
+
+    @staticmethod
+    def validate_dataclass_values(values: Dict):
+        for key, value in values.items():
+            if value is Ellipsis:
+                raise Exception(f"Missing required value for attribute: {key}")
+
 
 class MapScope(SFNScope):
     def __init__(self, parent_scope: SFNScope):
@@ -1103,14 +1401,15 @@ class MapScope(SFNScope):
     def _updated_var(self, var: str):
         self._updated_vars.add(var)
 
-    def build_entry_step(self, entry_var: str):
+    def build_entry_step(self, entry_var: str, index_var: str = None):
+        values = {entry_var: JsonPath.string_at(f"$.{entry_var}")}
+        if index_var:
+            values[index_var] = JsonPath.string_at(f"$.{index_var}")
         return sfn.Pass(
             self.cdk_stack,
             self.state_name("Register loop value"),
             result_path="$.register",
-            parameters=self.build_register_assignment(
-                {entry_var: JsonPath.string_at(f"$.{entry_var}")}, "register."
-            ),
+            parameters=self.build_register_assignment(values, "register."),
         )
 
     @property
@@ -1141,7 +1440,8 @@ def advance(
             chain = chain[0]
         if isinstance(next_, list):
             for i in flatten(next_):
-                i(chain)
+                if i and callable(i):
+                    i(chain)
         else:
             next_(chain)
         return new_next
@@ -1176,7 +1476,7 @@ def write_definition_json(name, start):
     with open(pathlib.Path("build", f"{name}.json"), "w") as fp:
         states = sfn.State.find_reachable_states(start, include_error_handlers=True)
         states.sort(
-            key=lambda state: int(state.id.split("[")[1].split("]")[0].split(":")[1])
+            key=lambda s: int(s.id[s.id.rindex("[") :].split("]")[0].split(":")[1])
         )
         json.dump(
             {"StartAt": start.id, "States": {s.id: s.to_state_json() for s in states},},
